import click
from cryptography.hazmat.primitives import serialization, hashes
from cryptography import x509

from ib1.directory.certificates import (
    create_signing_pair,
    load_certificate,
    load_key,
    build_subject,
    sign_csr,
    generate_key,
    get_bundle,
)


@click.group()
def cli():
    """A CLI for generating and signing certificates."""
    pass


@cli.command()
@click.option(
    "-u",
    "--usage",
    type=click.Choice(["signing", "client", "server"]),
    help="Choose signing, server or client CA",
    default="client",
)
@click.option(
    "-c", "--country", default="GB", help="Country to use for certificate generation"
)  # , state: str, framework: str
@click.option(
    "-s", "--state", default="London", help="State to use for certificate generation"
)
@click.option(
    "-f", "--framework", default="Core", help="Framework this certificate is for"
)
def create_ca(usage: str, country: str, state: str, framework: str):
    """Generate a server signing CA key and certificate and an issuer key and certificate pair signed by the CA then saves all files to disk"""
    print(f"Creating {usage} CA")
    ca_key, ca_certificate = create_signing_pair(
        country=country,
        state=state,
        framework=framework,
        use=usage.capitalize(),
        kind="CA",
    )
    issuer_key, issuer_certificate = create_signing_pair(
        country=country,
        state=state,
        framework=framework,
        use=usage.capitalize(),
        kind="Issuer",
        ca_cert=ca_certificate,
        ca_key=ca_key,
    )

    ca_certificate_name = f"{usage.lower()}-ca-cert.pem"
    with open(ca_certificate_name, "wb") as f:
        f.write(ca_certificate.public_bytes(serialization.Encoding.PEM))
    print(f"CA cert: {ca_certificate_name}")

    ca_key_name = f"{usage.lower()}-ca-key.pem"
    with open(ca_key_name, "wb") as f:
        f.write(
            ca_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )
    print(f"CA key: {ca_key_name}")

    issuer_certificate_name = f"{usage.lower()}-issuer-cert.pem"
    with open(issuer_certificate_name, "wb") as f:
        f.write(issuer_certificate.public_bytes(serialization.Encoding.PEM))
    print(f"Issuer cert: {issuer_certificate_name}")

    with open(f"{usage.lower()}-issuer-key.pem", "wb") as f:
        f.write(
            issuer_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )
    print(f"Issuer key: {usage.lower()}-issuer-key.pem")


@cli.command()
@click.option(
    "--issuer-key-file",
    type=click.File("rb"),
    help="Issuer key file",
    default="client-issuer-key.pem",
)
@click.option(
    "--issuer-cert-file",
    type=click.File("rb"),
    help="Issuer certificate file",
    default="client-issuer-cert.pem",
)
@click.option(
    "--member-uri",
    type=str,
    help="Member uri",
    default="https://directory.estf.ib1.org/member/2876152",
)
@click.option(
    "--application-uri",
    type=str,
    help="Application uri",
    default="https://directory.estf.ib1.org/scheme/electricty/application/26241",
)
@click.option(
    "--organization-name",
    type=str,
    help="Organization name",
    default="Demo Carbon Accounting Provider",
)
@click.option("--country", type=str, help="Country", default="GB")
@click.option("--state", type=str, help="State", default="London")
@click.option(
    "--role",
    "-r",
    help="Client roles",
    multiple=True,
    default=[
        "https://registry.estf.ib1.org/scheme/electricty/role/supply-voltage-reader"
    ],
)
def create_client_certificates(
    issuer_key_file: click.Path,
    issuer_cert_file: click.Path,
    member_uri: str,
    application_uri: str,
    organization_name: str,
    country: str,
    state: str,
    role: list[str],
):
    """
    Create a private key and use it generate a CSR, then sign the CSR with a CA key and certificate.

    Saves the private key, CSR, certificate and bundle to disk.
    """

    with open(issuer_cert_file.name, "rb") as f:
        issuer_cert_pem = f.read()
    with open(issuer_key_file.name, "rb") as f:
        issuer_key_pem = f.read()
    issuer_cert = load_certificate(issuer_cert_pem)
    issuer_key = load_key(issuer_key_pem)
    client_key = generate_key()
    subject = build_subject(
        country=country,
        state=state,
        organization_name=organization_name,
        common_name=application_uri,
    )
    csr = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(subject)
        .sign(client_key, hashes.SHA256())
    )
    csr_pem = csr.public_bytes(serialization.Encoding.PEM)
    # Create a CSR using the arguments given
    client_certificate = sign_csr(
        issuer_cert=issuer_cert,
        issuer_key=issuer_key,
        csr_pem=csr_pem,
        subject=subject,
        roles=role,
        member=member_uri,
    )
    client_certificate_pem = client_certificate.public_bytes(serialization.Encoding.PEM)
    bundle = get_bundle(client_certificate_pem, issuer_cert_pem)
    file_prefix = organization_name.lower().replace(" ", "-")
    # Write private key to disk as application-key.pem
    with open(f"{file_prefix}-key.pem", "wb") as f:
        f.write(
            client_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )
    # Write certifivate PEM to disk as application-cert.pem
    with open(f"{file_prefix}-cert.pem", "wb") as f:
        f.write(client_certificate_pem)
    # Write bundle to disk as application-bundle.pem
    with open(f"{file_prefix}-bundle.pem", "wb") as f:
        f.write(bundle)


@cli.command()
@click.option(
    "--issuer-key-file",
    type=click.File("rb"),
    help="Issuer key file",
    default="server-issuer-key.pem",
)
@click.option(
    "--issuer-cert-file",
    type=click.File("rb"),
    help="Issuer certificate file",
    default="server-issuer-cert.pem",
)
@click.option(
    "--domain",
    type=str,
    help="Domain name",
    default="http://tf-member.org",
)
@click.option(
    "--trust-framework",
    type=str,
    help="Trust framework",
    default="Core Trust Framework",
)
@click.option("--country", type=str, help="Country", default="GB")
@click.option("--state", type=str, help="State", default="London")
def create_server_certificates(
    issuer_key_file: click.Path,
    issuer_cert_file: click.Path,
    domain: str,
    trust_framework: str,
    country: str,
    state: str,
):
    """
    Create a private key and use it generate a CSR, then sign the CSR with a CA key and certificate.

    Saves the private key, CSR, certificate and bundle to disk.
    """

    with open(issuer_cert_file.name, "rb") as f:
        issuer_cert_pem = f.read()
    with open(issuer_key_file.name, "rb") as f:
        issuer_key_pem = f.read()
    issuer_cert = load_certificate(issuer_cert_pem)
    issuer_key = load_key(issuer_key_pem)
    server_key = generate_key()
    subject = build_subject(
        country=country,
        state=state,
        organization_name=trust_framework,
        common_name=domain,
    )
    csr = (
        x509.CertificateSigningRequestBuilder()
        .subject_name(subject)
        .sign(server_key, hashes.SHA256())
    )
    csr_pem = csr.public_bytes(serialization.Encoding.PEM)
    # Create a CSR using the arguments given
    server_certificate = sign_csr(
        issuer_cert=issuer_cert,
        issuer_key=issuer_key,
        csr_pem=csr_pem,
        subject=subject,
        server=True,
    )
    server_certificate_pem = server_certificate.public_bytes(serialization.Encoding.PEM)
    bundle = get_bundle(server_certificate_pem, issuer_cert_pem)
    file_prefix = domain.replace("/", "").replace(":", "")
    # Write private key to disk as application-key.pem
    with open(f"{file_prefix}-key.pem", "wb") as f:
        f.write(
            server_key.private_bytes(
                encoding=serialization.Encoding.PEM,
                format=serialization.PrivateFormat.TraditionalOpenSSL,
                encryption_algorithm=serialization.NoEncryption(),
            )
        )
    # Write certifivate PEM to disk as application-cert.pem
    with open(f"{file_prefix}-cert.pem", "wb") as f:
        f.write(server_certificate_pem)
    # Write bundle to disk as application-bundle.pem
    with open(f"{file_prefix}-bundle.pem", "wb") as f:
        f.write(bundle)


if __name__ == "__main__":
    cli()



from urllib.parse import unquote
from cryptography import x509
from cryptography.hazmat.backends import default_backend

from ib1.directory.exceptions import CertificateInvalidError


def parse_cert(client_certificate: str) -> x509.Certificate:
    """
    Given a certificate as a quoted or unquoted string, return a x509.Certificate object.

    Parameters:
        client_certificate (str): The client certificate as a string, which can be either quoted or unquoted.

    Returns:
        x509.Certificate: The parsed x509.Certificate object.

    Raises:
        CertificateInvalidError: If the certificate string is invalid and cannot be parsed.
    """
    try:
        return x509.load_pem_x509_certificate(
            bytes(unquote(client_certificate), "utf-8"), default_backend()
        )
    except ValueError:
        raise CertificateInvalidError("Invalid certificate string")



from asn1crypto.core import SequenceOf, UTF8String  # type: ignore


class UTF8Sequence(SequenceOf):
    _child_spec = UTF8String


def decode_sequence(der_bytes: bytes) -> list[str]:
    values = UTF8Sequence.load(der_bytes)
    decoded = []
    for i in range(0, len(values)):
        decoded.append(values[i].native)
    return decoded


def encode_sequence(urls: list[str]) -> bytes:
    extension_sequence = UTF8Sequence(urls)
    return extension_sequence.dump()


def encode_string(string: str) -> bytes:
    return UTF8String(string).dump()


def decode_string(der_bytes: bytes) -> str:
    return str(UTF8String.load(der_bytes))



from .exceptions import (
    CertificateError,
    CertificateMissingError,
    CertificateInvalidError,
    CertificateExtensionError,
    CertificateRoleError,
)

from . import certificates

from .extensions import require_role
from .utils import parse_cert

__all__ = [
    "CertificateError",
    "CertificateMissingError",
    "CertificateInvalidError",
    "CertificateExtensionError",
    "CertificateRoleError",
    "require_role",
    "parse_cert",
    "certificates",
]



class Ib1DirectoryError(Exception):
    """Base class for exceptions in this module."""


class CertificateError(Ib1DirectoryError):
    """
    Base class for errors related to the client certificate
    """


class CertificateMissingError(CertificateError):
    """
    Raised when the client certificate is missing
    """


class CertificateInvalidError(CertificateError):
    pass


class CertificateExtensionError(CertificateError):
    pass


class CertificateRoleError(CertificateError):
    pass



import datetime
from typing import Tuple, List
from io import BytesIO

from cryptography import x509
from cryptography.hazmat.backends import default_backend
from cryptography.x509.oid import NameOID
from cryptography.hazmat.primitives.asymmetric import ec
from cryptography.hazmat.primitives import hashes, serialization
from cryptography.hazmat.primitives.asymmetric.types import PrivateKeyTypes

from ib1.directory.extensions import encode_roles, encode_member


def _ca_extensions_cert(
    subject: x509.Name,
    issuer_name: x509.Name,
    issuer_key: ec.EllipticCurvePrivateKey,
    signing_key: ec.EllipticCurvePrivateKey,
    ca_path_length: int | None = 0,
    valid_from: datetime = datetime.datetime.now(datetime.timezone.utc),
    valid_to: datetime = datetime.datetime.now(datetime.timezone.utc)
    + datetime.timedelta(days=365),
) -> x509.Certificate:
    builder = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer_name)
        .public_key(issuer_key.public_key())
        .serial_number(x509.random_serial_number())
        .not_valid_before(valid_from)
        .not_valid_after(valid_to)
        .add_extension(
            x509.SubjectKeyIdentifier.from_public_key(issuer_key.public_key()),
            critical=False,
        )
        .add_extension(
            x509.AuthorityKeyIdentifier.from_issuer_public_key(
                signing_key.public_key()
            ),
            critical=False,
        )
        .add_extension(
            x509.BasicConstraints(ca=True, path_length=ca_path_length), critical=True
        )
        .add_extension(
            x509.KeyUsage(
                digital_signature=True,
                key_cert_sign=True,
                crl_sign=True,
                key_agreement=False,
                content_commitment=False,
                encipher_only=False,
                decipher_only=False,
                key_encipherment=False,
                data_encipherment=False,
            ),
            critical=True,
        )
    )

    return builder.sign(signing_key, hashes.SHA256(), default_backend())


def load_certificate(cert_pem: bytes) -> x509.Certificate:
    """
    Load a certificate from a PEM-encoded byte string.

    Args:
        cert_pem (bytes): PEM-encoded certificate.

    Returns:
        x509.Certificate: The loaded certificate.
    """
    return x509.load_pem_x509_certificate(cert_pem, default_backend())


def load_key(key_pem: bytes) -> PrivateKeyTypes:
    """
    Load a private key from a PEM-encoded byte string.

    Args:
        key_pem (bytes): PEM-encoded private key.

    Returns:
        PrivateKeyTypes: The loaded private key.
    """
    return serialization.load_pem_private_key(
        key_pem, password=None, backend=default_backend()
    )


def build_subject(
    country: str,
    state: str,
    organization_name: str,
    common_name: str,
) -> x509.Name:
    """
    Build an X.509 subject name.

    Args:
        country (str): Country name.
        state (str): State or province name.
        organization_name (str): Organization name.
        common_name (str): Common name.

    Returns:
        x509.Name: The constructed X.509 subject name.
    """
    return x509.Name(
        [
            x509.NameAttribute(NameOID.COUNTRY_NAME, country),
            x509.NameAttribute(NameOID.STATE_OR_PROVINCE_NAME, state),
            x509.NameAttribute(NameOID.ORGANIZATION_NAME, organization_name),
            x509.NameAttribute(NameOID.COMMON_NAME, common_name),
        ]
    )


def generate_key():
    return ec.generate_private_key(ec.SECP256R1(), default_backend())


def generate_root_key():
    return ec.generate_private_key(ec.SECP384R1(), default_backend())


def create_signing_pair(
    country: str = "GB",
    state: str = "London",
    framework: str = "Core",
    use: str = "Client",
    kind: str = "CA",
    ca_cert: x509.Certificate | None = None,
    ca_key: ec.EllipticCurvePrivateKey | None = None,
) -> Tuple[ec.EllipticCurvePrivateKey, x509.Certificate]:
    """
    Create a signing key certificate pair.

    Args:
        country (str): Country name.
        state (str): State or province name.
        framework (str): Framework name.
        use (str): Use case (e.g., Client, Server).
        kind (str): Kind of certificate (e.g., CA).
        ca_cert (x509.Certificate, optional): CA certificate.
        ca_key (ec.EllipticCurvePrivateKey, optional): CA private key.

    Returns:
        Tuple[ec.EllipticCurvePrivateKey, x509.Certificate]: The generated private key and certificate.
    """
    if kind == "CA":
        key = generate_root_key()
        valid_to = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(
            days=9132
        )
    else:
        key = generate_key()
        valid_to = datetime.datetime.now(datetime.timezone.utc) + datetime.timedelta(
            days=365
        )
    description = f"{framework} Trust Framework"
    subject_name = build_subject(
        country, state, description, f"{description} {use} {kind}"
    )
    if ca_cert is None or ca_key is None:
        issuer = subject_name
        ca_key = key
        path_length = None
    else:
        issuer = ca_cert.subject
        path_length = 0

    # Build the certificate
    if ca_key:
        cert = _ca_extensions_cert(
            subject=subject_name,
            issuer_name=issuer,
            issuer_key=key,
            signing_key=ca_key,
            ca_path_length=path_length,
            valid_to=valid_to,
        )
    return key, cert


def get_bundle(certificate_pem: bytes, issuer_cert_pem: bytes) -> bytes:

    # Concatenate certificates in memory
    bundle = BytesIO()
    bundle.write(certificate_pem)
    bundle.write(issuer_cert_pem)

    # Get the concatenated bundle as bytes
    bundle_pem = bundle.getvalue()

    # Close the BytesIO object
    bundle.close()
    return bundle_pem


def sign_csr(
    issuer_cert: x509.Certificate,
    issuer_key: ec.EllipticCurvePrivateKey,
    csr_pem: bytes,
    subject: x509.Name,
    roles: List[str] | None = None,
    member: str | None = None,
    server: bool = False,
    days_valid: int = 365,
) -> x509.Certificate:
    """
    Sign a user-provided CSR.

    Args:
        issuer_cert (x509.Certificate): Issuer certificate.
        issuer_key (ec.EllipticCurvePrivateKey): Issuer private key.
        csr_pem (bytes): CSR in PEM format.
        days_valid (int): Number of days the certificate is valid for.
        subject (x509.Name): Subject name for the certificate.
        roles (List[str], optional): Roles to encode in the certificate as a custom extension
        member (str, optional): Member to encode in the certificate as a custom extension
        server (bool, optional): Whether the certificate is for a server.

    Returns:
        x509.Certificate: The signed certificate.
    """
    csr = x509.load_pem_x509_csr(csr_pem, default_backend())

    # Build the application certificate
    cert_builder = (
        x509.CertificateBuilder()
        .subject_name(subject)
        .issuer_name(issuer_cert.subject)
        .public_key(csr.public_key())
        .not_valid_before(datetime.datetime.now(datetime.timezone.utc))
        .not_valid_after(
            datetime.datetime.now(datetime.timezone.utc)
            + datetime.timedelta(days=days_valid)
        )
        .serial_number(x509.random_serial_number())
    )
    # Add SKI (Subject Key Identifier)
    subject_ski = x509.SubjectKeyIdentifier.from_public_key(csr.public_key())
    cert_builder = cert_builder.add_extension(subject_ski, critical=False)

    # Add AKI (Authority Key Identifier)
    issuer_public_key = issuer_cert.public_key()
    authority_aki = x509.AuthorityKeyIdentifier.from_issuer_public_key(
        issuer_public_key
    )
    cert_builder = cert_builder.add_extension(authority_aki, critical=False)

    if server:
        cert_builder = cert_builder.add_extension(
            x509.ExtendedKeyUsage([x509.oid.ExtendedKeyUsageOID.SERVER_AUTH]),
            critical=True,
        )
        cert_builder = cert_builder.add_extension(
            x509.KeyUsage(
                digital_signature=True,
                key_cert_sign=False,
                crl_sign=False,
                key_agreement=False,
                content_commitment=False,
                encipher_only=False,
                decipher_only=False,
                key_encipherment=True,
                data_encipherment=False,
            ),
            critical=True,
        )
    else:
        common_name = str(subject.get_attributes_for_oid(NameOID.COMMON_NAME)[0].value)
        cert_builder = cert_builder.add_extension(
            x509.SubjectAlternativeName([x509.UniformResourceIdentifier(common_name)]),
            critical=False,
        )

    if roles:
        cert_builder = encode_roles(cert_builder, roles)
    if member:
        cert_builder = encode_member(cert_builder, member)

    cert = cert_builder.sign(issuer_key, hashes.SHA256(), default_backend())
    return cert



from cryptography import x509

from ib1.directory.exceptions import CertificateRoleError, CertificateExtensionError
from ib1.directory import der

ROLE_IDENTIFIER = "1.3.6.1.4.1.62329.1.1"
MEMBER_IDENTIFIER = "1.3.6.1.4.1.62329.1.3"


def _add_extension(
    cert_builder: x509.CertificateBuilder, oid: str, value: bytes
) -> x509.CertificateBuilder:
    """
    Add an extension to the certificate builder.

    Args:
        cert_builder (x509.CertificateBuilder): The certificate builder.
        oid (str): The object identifier for the extension.
        value (bytes): The value of the extension.

    Returns:
        x509.CertificateBuilder: The updated certificate builder with the new extension.
    """
    return cert_builder.add_extension(
        x509.UnrecognizedExtension(
            x509.ObjectIdentifier(oid),
            value,
        ),
        critical=False,
    )


def _extension_value(cert: x509.Certificate, oid: str) -> bytes:
    """
    Retrieve the value of an extension from a certificate.

    Args:
        cert (x509.Certificate): The certificate.
        oid (str): The object identifier for the extension.

    Returns:
        bytes: The value of the extension.
    """
    try:
        extension = cert.extensions.get_extension_for_oid(
            x509.ObjectIdentifier(oid)
        ).value.value  # type: ignore [attr-defined]
    except AttributeError:
        raise CertificateExtensionError("Certificate is invalid or missing extensions")
    return extension


def encode_roles(cert_builder: x509.CertificateBuilder, roles: list[str]):
    """
    Encode roles into the certificate builder as an extension.

    Args:
        cert_builder (x509.CertificateBuilder): The certificate builder.
        roles (list[str]): The roles to encode.

    Returns:
        x509.CertificateBuilder: The updated certificate builder with the roles extension.
    """
    return _add_extension(cert_builder, ROLE_IDENTIFIER, der.encode_sequence(roles))


def encode_member(cert_builder: x509.CertificateBuilder, member: str):
    """
    Encode member information into the certificate builder as an extension.

    Args:
        cert_builder (x509.CertificateBuilder): The certificate builder.
        member (str): The member information to encode.

    Returns:
        x509.CertificateBuilder: The updated certificate builder with the member extension.
    """
    return _add_extension(cert_builder, MEMBER_IDENTIFIER, der.encode_string(member))


def decode_roles(cert: x509.Certificate) -> list[str]:
    """
    Decode roles from a certificate.

    Args:
        cert (x509.Certificate): The certificate.

    Returns:
        list[str]: The decoded roles.

    Raises:
        CertificateExtensionError: If the certificate does not include role information.
    """
    try:
        role_der = _extension_value(cert, ROLE_IDENTIFIER)
    except x509.ExtensionNotFound:
        raise CertificateExtensionError(
            "Client certificate does not include role information"
        )
    return der.decode_sequence(role_der)


def decode_member(cert: x509.Certificate) -> str:
    """
    Decode member information from a certificate.

    Args:
        cert (x509.Certificate): The certificate.

    Returns:
        str: The decoded member information.

    Raises:
        CertificateExtensionError: If the certificate does not include member information.
    """
    try:
        member_der = _extension_value(cert, MEMBER_IDENTIFIER)
    except x509.ExtensionNotFound:
        raise CertificateExtensionError(
            "Client certificate does not include member information"
        )
    return der.decode_string(member_der)


def require_role(role_name: str, cert: x509.Certificate) -> bool:
    """
    Check that the certificate includes the given role, raising an exception if not.

    Args:
        role_name (str): The role name to check for.
        cert (x509.Certificate): The certificate.

    Returns:
        bool: True if the role is present in the certificate.

    Raises:
        CertificateRoleError: If the certificate does not include the role or the role information.
    """
    roles = decode_roles(cert)
    if role_name not in roles:
        raise CertificateRoleError(
            "Client certificate does not include role " + role_name
        )
    return True
